#include "stdafx.h"
#include "../../../code/vc6/alpha1/src/blepo.h"
using namespace blepo;
#include "match_vnk.h"

MatchVNK::MatchVNK(const char* path_ref_img, const char* path_warp_img)
{

  ImgGray ref_img_gray, warp_img_gray;
  Load(path_ref_img, &ref_img_gray);
  Load(path_warp_img, &warp_img_gray);

  Convert(ref_img_gray, &mk_ref_img);
  Convert(warp_img_gray, &mk_warp_img);

  assert(mk_ref_img.Width() == mk_warp_img.Width());
  assert(mk_ref_img.Height() == mk_warp_img.Height());

  mk_width = mk_ref_img.Width();
  mk_height = mk_ref_img.Height();

  mk_ref_min_img.Reset(mk_width, mk_height);
  mk_ref_max_img.Reset(mk_width, mk_height);
  mk_warp_min_img.Reset(mk_width, mk_height);
  mk_warp_max_img.Reset(mk_width, mk_height);

  // initialising label image
  ResetLabelImage();

  // initialising data_cost image
  mk_data_cost_img.Reset(mk_width, mk_height);
  Set(&mk_data_cost_img, 0);
  
  mk_ptr_img = new PtrRaw[mk_width*mk_height];
  Location p;
	  for (p.t=0; p.t<=0; p.t++)
	    for (p.y=0; p.y<mk_height; p.y++)
	      for (p.x=0; p.x<mk_width; p.x++)
	        *(mk_ptr_img + ((p.y)*mk_width + p.x)) = mk_ptr_img;

  // intialising correlation related images
  mk_corr_label.Reset(mk_width, mk_height);
  Set(&mk_corr_label, 0);

  mk_corr_hard.Reset(mk_width, mk_height);
  Set(&mk_corr_hard, 0);

  N[0]= Location(-1,0,0);
	N[1]= Location(1,0,0);
	N[2]= Location(0,-1,0);
	N[3]= Location(0,1,0);

	PS = Location(0,0,0);
	PE = Location(mk_width-1, mk_height-1, 0);

  mk_disp_x_min = mvnk_disp_x_min;
  mk_disp_x_max = mvnk_disp_x_max;
  mk_disp_y_min = mvnk_disp_y_min;
  mk_disp_y_max = mvnk_disp_y_max;

  mk_small_region_ratio = small_region_ratio;
  mk_small_region_area = mk_width*mk_height*mk_small_region_ratio;

  ConstructMinMaxImages();
  InitLabels();

  mk_multiwaycut_time = 0.0;
  mk_affine_time = 0.0;
  mk_oversegmentation_time = 0.0;
  mk_total_time = 0.0;

  mk_main_iter_num = 0;

  // finding hard constraint points
  // works only for stereo
  #ifdef HARD_CONSTRAINTS
    Correlation();
    mk_label = mk_corr_label;
  #endif
  
  #ifdef CANNY_SMOOTHNESS
    ImgBinary edge;
    Canny(ref_img_gray, 1.0f, &edge, 0.8f, 5.0f);
    Convert(edge, &mk_canny_edge);
  #endif
  //ImgGray hard_constraints;
  //RealTimeStereo(ref_img_gray,warp_img_gray,&hard_constraints,(mk_disp_x_max-mk_disp_x_min));
  //Load("C:/Research/Cppexp/Learning/match_corr_vnk/output/hard_disp.pgm",&hard_constraints);
  //Convert(hard_constraints, &mk_hard_constraints);

  /*Figure fig;
  fig.Draw(mk_hard_constraints);
  AfxMessageBox("ok");*/

  //AfxMessageBox("MatchVNK object constructed");
}

MatchVNK::MatchVNK()
{
  AfxMessageBox("Warning: Empty match_vnk construction");
}

MatchVNK::~MatchVNK()
{
  if(mk_ptr_img)
    delete[] mk_ptr_img;

  //AfxMessageBox("MatchVNK object destructed");
}

void MatchVNK::ConstructMinMaxImages()
{
  Location p;
	float I, I1, I2, I3, I4, I_min, I_max;

  // reference image min and max images
  for (p.t=0; p.t<=0; p.t++)
	  for (p.y=0; p.y<mk_height; p.y++)
	    for (p.x=0; p.x<mk_width; p.x++)
	    {
		    I = I_min = I_max = (*(mk_ref_img.Begin(p.x,p.y)));

		    if (p.x>0)           I1 = ((*(mk_ref_img.Begin(p.x-1,p.y))) + I) / 2;
		    else                 I1 = I;
		    if (p.x<mk_width-1)  I2 = ((*(mk_ref_img.Begin(p.x+1,p.y))) + I) / 2;
		    else                 I2 = I;
		    if (p.y>0)           I3 = ((*(mk_ref_img.Begin(p.x,p.y-1))) + I) / 2;
		    else                 I3 = I;
		    if (p.y<mk_height-1) I4 = ((*(mk_ref_img.Begin(p.x,p.y+1))) + I) / 2;
		    else                 I4 = I;

		    if (I_min > I1) I_min = I1;
		    if (I_min > I2) I_min = I2;
		    if (I_min > I3) I_min = I3;
		    if (I_min > I4) I_min = I4;
		    if (I_max < I1) I_max = I1;
		    if (I_max < I2) I_max = I2;
		    if (I_max < I3) I_max = I3;
		    if (I_max < I4) I_max = I4;

		    (*(mk_ref_min_img.Begin(p.x,p.y))) = I_min;
		    (*(mk_ref_max_img.Begin(p.x,p.y))) = I_max;
      }

  // warp image min and max images
  for (p.t=0; p.t<=0; p.t++)
	  for (p.y=0; p.y<mk_height; p.y++)
	    for (p.x=0; p.x<mk_width; p.x++)
	    {
		    I = I_min = I_max = (*(mk_warp_img.Begin(p.x,p.y)));

		    if (p.x>0)           I1 = ((*(mk_warp_img.Begin(p.x-1,p.y))) + I) / 2;
		    else                 I1 = I;
		    if (p.x<mk_width-1)  I2 = ((*(mk_warp_img.Begin(p.x+1,p.y))) + I) / 2;
		    else                 I2 = I;
		    if (p.y>0)           I3 = ((*(mk_warp_img.Begin(p.x,p.y-1))) + I) / 2;
		    else                 I3 = I;
		    if (p.y<mk_height-1) I4 = ((*(mk_warp_img.Begin(p.x,p.y+1))) + I) / 2;
		    else                 I4 = I;

		    if (I_min > I1) I_min = I1;
		    if (I_min > I2) I_min = I2;
		    if (I_min > I3) I_min = I3;
		    if (I_min > I4) I_min = I4;
		    if (I_max < I1) I_max = I1;
		    if (I_max < I2) I_max = I2;
		    if (I_max < I3) I_max = I3;
		    if (I_max < I4) I_max = I4;

		    (*(mk_warp_min_img.Begin(p.x,p.y))) = I_min;
		    (*(mk_warp_max_img.Begin(p.x,p.y))) = I_max;
      }
}

void MatchVNK::RunTimeInfo()
{
  char buffer[500];
  
  sprintf(buffer, "number of iterations is %d\n multiway cut time is %lf\n \
                   affine time is %lf\n oversegmentation time is %lf\n \
                   total time is %lf\n", mk_main_iter_num, mk_multiwaycut_time,
                   mk_affine_time, mk_oversegmentation_time, mk_total_time);

  AfxMessageBox(buffer);
}

