// match_vnk.h
// Implementation of multiwaycut affine work of S. Birchfield and C.Tomasi, ICCV 99
// Braga Natarajan
// date started: February 2005
// date updated: December 2005

// some parts of codes in this project are based on:
// MATCH - software for computing dense correspondence     
// (disparity map) between two images using graph cuts
// Version 3.1
// http://www.cs.cornell.edu/People/vnk/recon.html
// Vladimir N. Kolmogorov
// vnk@cs.cornell.edu 2001-2003

//#define HARD_CONSTRAINTS
//#define CANNY_SMOOTHNESS
//#define SAVE_CORR_VECTOR

const int INFINITY = 10000;

const float data_cost_limit = 1000.0f;
const float data_cost_oob = 5.0f;

const int mvnk_disp_x_min = -30;
const int mvnk_disp_x_max = 0;
const int mvnk_disp_y_min = 0;
const int mvnk_disp_y_max = 0;

const float small_region_ratio = 0.009f;

const int corr_size = 5;
const int default_v = 100;

const int alpha = 200;
const float beta1 = 0.2f;
const float beta2 = 0.2f;
const float gamma = 0.4f;
const int hard_area_limit = 60;

const float disparity_scale_factor = 8.0f;
const int hard_points_scale_factor = 8;

const int mvnk_iter_num = 1;

typedef void* PtrRaw;

struct Location 
{
	
	int x, y, t;

	Location() {}
	Location(int a, int b, int c) { x = a; y = b; t=c; }

	Location operator- ()        { return Location(-x, -y, -t); }
	Location operator+ (Location a) { return Location(x + a.x, y + a.y, t + a.t); }
	Location operator- (Location a) { return Location(x - a.x, y - a.y, t - a.t); }

	bool  operator< (Location a) { return (x <  a.x) && (y <  a.y) && (t < a.t); }
	bool  operator<=(Location a) { return (x <= a.x) && (y <= a.y) && (t <= a.t); }
	bool  operator> (Location a) { return (x >  a.x) && (y >  a.y) && (t > a.t); }
	bool  operator>=(Location a) { return (x >= a.x) && (y >= a.y) && (t >= a.t); }
	bool  operator==(Location a) { return (x == a.x) && (y == a.y) && (t == a.t); }
	bool  operator!=(Location a) { return (x != a.x) || (y != a.y) || (t != a.t); }
};

class MatchVNK
{
private:

  int mk_width;
  int mk_height;

  ImgFloat mk_ref_img;
  ImgFloat mk_warp_img;

  ImgFloat mk_ref_min_img;
  ImgFloat mk_ref_max_img;
  ImgFloat mk_warp_min_img;
  ImgFloat mk_warp_max_img;

  ImgInt mk_label;
  ImgInt mk_stb_label_reconnect;
  ImgFloat mk_disp_map;

  ImgFloat mk_hard_constraints; // by RealTimeStereo

  ImgInt mk_data_cost_img;

  PtrRaw *mk_ptr_img; 

  ImgInt mk_corr_label;
  ImgInt mk_corr_hard;

  ImgInt mk_canny_edge;

  Location N[4];
  Location PS;
  Location PE;

  int mk_disp_x_min;
  int mk_disp_x_max;
  int mk_disp_y_min;
  int mk_disp_y_max;

  float mk_small_region_ratio;
  int mk_small_region_area;

  std::vector<MatFlt> mk_stb_disp_func;
  std::vector<float> mk_stb_centroid_x;
  std::vector<float> mk_stb_centroid_y;

  std::vector<int> mk_disp_x;
  std::vector<int> mk_disp_y;

  double mk_multiwaycut_time;
  double mk_affine_time;
  double mk_oversegmentation_time;
  double mk_total_time;

  int mk_main_iter_num;

// private functions

  void CorrHor(int disp_x, int disp_y, ImgInt& v_hor);
  void CorrFull(const ImgInt& v_hor, ImgInt& v_full);
  void CorrelationLR(bool left_reference);
  void Correlation();  
  int DataCostCorr(Location p, Location p_disp);
  void ScaleHardPoints(ImgGray* out);

  void ConstructMinMaxImages();
  void InitLabels();
  void ResetLabelImage();
  void OverSegmentation();
  void CreateDispMap();
  void AdjustBorder();
  
  int GetEnergy();
  int GetChangedEnergy(const int label1, const int label2);
  void AExpand(int label);
  void MSegment();

  void PredictOcclusion();

  void MatchVNK::OcclusionAffineWarp(
    const int val1,
    const int val2,
    std::vector<Point>* loc1_occlusion,
    std::vector<Point>* loc2_occlusion,
    std::vector<Point>* loc1_opp_occlusion,
    std::vector<Point>* loc2_opp_occlusion);

  bool FloatPointVal(
    const float xf, 
    const float yf,
    const int val,
    int* xo,
    int* yo);

  void MatchVNK::NeighborRegionBoundary(
    const int val1,
    const int val2,
    std::vector<Point>* boundary1,
    std::vector<Point>* boundary2);

  int __cdecl DataCostInit(Location p, int p_lbl);
  int __cdecl DataCostIter(Location p, int p_lbl);
  int (__cdecl MatchVNK::*DataCost)(Location p, int p_lbl);
  int SmoothnessCost(Location p, int p_lbl, Location q, int q_lbl);

public:

// variables
  int E;

// functions

  MatchVNK(const char* path_ref_img, const char* path_warp_img);
  MatchVNK();
  ~MatchVNK();

  void IterSegment(bool init);
  void SaveLabelImage(char* file_name);
  void DisplayLabelImage();
  void DebugFunc(); // function to test internal working
  void RunTimeInfo();
};
  