#include "stdafx.h"
#include "../../../code/vc6/alpha1/src/blepo.h"
#include "../LogTrace/Trace.h"
using namespace blepo;
#include "match_vnk.h"
#include "energy.h"
#include "diffuse.h"
#include "bnutility.h"

const float I_threshold = 5.0f;
const int lambda1 = 12;
const int lambda2 = 6;

int multiway_cut_iter_num;

const float hard_penalty_factor = 6.0f;

MatFlt A(2,2), d(1,2), centroid(1,2);
MatFlt loc_in(1,2), loc_out(1,2);

LKAParameters params;

const bool do_oversegm = true;

const bool skip_label_zero = false;
const bool skip_label_zero_region = false;
const bool skip_oversegm_label_zero  = false;

const int border_adjust_width = 20;

void MatchVNK::IterSegment(bool init)
{
  clock_t start, stop, begin, end;

  begin = clock();

  ImgInt label_tmp;
  std::vector<MatFlt> disp_func_tmp;
  std::vector<float> centroid_x_tmp;
  std::vector<float> centroid_y_tmp;

  DataCost = DataCostInit;
  multiway_cut_iter_num = 2;
  int E_cur, E_prev;
  int iter=0;
  while(1)
  {
    //Trace("\nMain iteration loop number %d\n",iter);
    //Trace("------------------------------------------\n");
    
    char path_label[200];
    sprintf(path_label,"../../output/label%02d.pgm",iter);

    start = clock();
    if((init == false) && (iter == 0))
    {
      ImgGray label_gray;
      Load(path_label,&label_gray);
      Convert(label_gray,&mk_label);
    }
    else
    {
      //ResetLabelImage();
      //Figure fig1;
      //fig1.Draw(mk_label);
      //AfxMessageBox("ok");

      MSegment();
      //Trace("\nIterSegment loop number %d is done\n",iter);
      SaveLabelImage(path_label);
    }
    stop = clock();
    mk_multiwaycut_time += (((double)(stop - start))/CLOCKS_PER_SEC);

    start = clock();

    DataCost = DataCostIter;
    multiway_cut_iter_num = 1;

    AdjustBorder();

    //ImgInt label_smooth;
    //MedianFilterFinal(mk_label,3,3,&label_smooth);

    ImgInt label_diffused, label_diffused_original;
    std::vector<int> active_labels;
    //DiffuseRegions(mk_label, mk_small_region_area, &label_diffused, 
                   //&label_diffused_original, &active_labels);
    CoerceRegions(mk_label, mk_small_region_area, &label_diffused, 
                  &label_diffused_original, &active_labels);
    //Trace("max label in diffused image is %d\n",Max(label_diffused));
    //Trace("size of active_labels is %d\n",active_labels.size());

    // saving temporary images
    /*ImgGray label_diffused_gray;
    Convert(label_diffused,&label_diffused_gray);
    char path_label_diffused[200];
    sprintf(path_label_diffused,"../../output/label_diffused%02d.pgm",iter);
    Save(label_diffused_gray,path_label_diffused,"pgm");*/
    //

    mk_stb_label_reconnect = label_diffused;

    int min_label = Min(mk_stb_label_reconnect);
    int max_label = Max(mk_stb_label_reconnect);

    if(max_label > 100)
      AfxMessageBox("Warning: labels going beyond 100\n");
  
    //Trace("min is %d and max is %d\n",min_label,max_label);
    //TraceIntArray(active_labels);

    std::vector<MatFlt> disp_func_tmp = mk_stb_disp_func;
    std::vector<float> centroid_x_tmp = mk_stb_centroid_x;
    std::vector<float> centroid_y_tmp = mk_stb_centroid_y;

    mk_stb_disp_func.resize(max_label+1);
    mk_stb_centroid_x.resize(max_label+1);
    mk_stb_centroid_y.resize(max_label+1);

    // label correspondence between previous label image and new diffused image
    int new_label, old_label;
    for(new_label=min_label; new_label<=max_label; new_label++)
    {
      old_label = active_labels[new_label];
      mk_stb_disp_func[new_label] = disp_func_tmp[old_label];
      mk_stb_centroid_x[new_label] = centroid_x_tmp[old_label];
      mk_stb_centroid_y[new_label] = centroid_y_tmp[old_label];
    }

    // debug1 file - writes disp function after diffusing stage
    std::vector<int> sts;
    sts.resize(max_label+1,9);
    char* debug_file_path1 = "../../output/debug1.txt";
    //WriteResults(mk_stb_disp_func,mk_stb_centroid_x,mk_stb_centroid_y,sts,debug_file_path1);
    Trace("labels after coercion\n");
    TraceDispFunc(mk_stb_disp_func,mk_stb_centroid_x,mk_stb_centroid_y);
    //

    std::vector<MatFlt> A_init_regions;
    std::vector<MatFlt> d_init_regions;    
    if(skip_label_zero) min_label++;
    for(new_label=min_label; new_label<=max_label; new_label++)
    {
      MatFlt z;
      z = mk_stb_disp_func[new_label];

      MatFlt A_init(2,2);
      MatFlt d_init(1,2);

      A_init(0,0) = z(0,0);
      A_init(1,0) = z(1,0);
      d_init(0,0) = z(2,0);
      A_init(0,1) = z(3,0);
      A_init(1,1) = z(4,0);
      d_init(0,1) = z(5,0);

      A_init_regions.push_back(A_init);
      d_init_regions.push_back(d_init);
    }

    std::vector<LKAResult> result_regions;
    params.skip_label_zero = skip_label_zero_region;
    LKARegions(mk_ref_img, mk_warp_img, mk_stb_label_reconnect, params, 
               A_init_regions, d_init_regions, &result_regions);
  
    // debug2 file - writes affine parameters for each label
    char* debug_file_path2 = "../../output/debug2.txt";
    //WriteResults(result_regions, debug_file_path2);
    Trace("result of affine estimation\n");
    TraceLKAResultArray(result_regions);
    //    

    // IMPORTANT - Label number 0 condition
    std::vector<LKAResult>::const_iterator rr_p;
    int label_aff;
    label_aff = (skip_label_zero ? 1 : 0);
    for(rr_p=result_regions.begin(); rr_p!=result_regions.end(); rr_p++)
    {
      if(rr_p->status == 0) // if tracked
      {
        MatFlt z(6,1);

        z(0,0) = rr_p->A(0,0);
        z(1,0) = rr_p->A(1,0);
        z(2,0) = rr_p->d(0,0);
        z(3,0) = rr_p->A(0,1);
        z(4,0) = rr_p->A(1,1);
        z(5,0) = rr_p->d(0,1);

        mk_stb_disp_func[label_aff] = z;
        mk_stb_centroid_x[label_aff] = rr_p->centroid_x;
        mk_stb_centroid_y[label_aff] = rr_p->centroid_y;
      }
      label_aff++;
    }

    // debug3 file - writes updated displacement function after LucasKanadeAffine step
    sts.resize(max_label+1,9);
    char* debug_file_path3 = "../../output/debug3.txt";
    //WriteResults(mk_stb_disp_func,mk_stb_centroid_x,mk_stb_centroid_y,sts,debug_file_path3);
    Trace("labels after affine updating\n"); 
    TraceDispFunc(mk_stb_disp_func,mk_stb_centroid_x,mk_stb_centroid_y);
    //

    mk_label = mk_stb_label_reconnect;

    E_cur = GetEnergy();
    if(iter==0)
    {
      E_prev = E_cur;
      label_tmp = mk_label;
      disp_func_tmp = mk_stb_disp_func;
      centroid_x_tmp = mk_stb_centroid_x;
      centroid_y_tmp = mk_stb_centroid_y;
    }

    //Trace("\nenergy at the end of loop is %d\n",E_cur);

    if(E_cur > E_prev)
    {
      mk_label = label_tmp;
      mk_stb_disp_func = disp_func_tmp;
      mk_stb_centroid_x = centroid_x_tmp;
      mk_stb_centroid_y = centroid_y_tmp;
      break;
    }

    E_prev = E_cur;
    label_tmp = mk_label;
    disp_func_tmp = mk_stb_disp_func;
    centroid_x_tmp = mk_stb_centroid_x;
    centroid_y_tmp = mk_stb_centroid_y;
    iter++;
    
    stop = clock();
    mk_affine_time += (((double)(stop - start))/CLOCKS_PER_SEC);
  }

  mk_main_iter_num = iter+1;

  start = clock();
  E_cur = GetEnergy();
  E_prev = E_cur+1;
  while(E_cur < E_prev)
  {
    OverSegmentation();
    E_prev = E_cur;
    E_cur = GetEnergy();
  }
  stop = clock();
  mk_oversegmentation_time += (((double)(stop - start))/CLOCKS_PER_SEC);

  char path_label_oversegm[200];
  sprintf(path_label_oversegm,"../../output/label_oversegm.pgm");
  SaveLabelImage(path_label_oversegm);

  CreateDispMap();
  ImgGray disp_map_gray, disp_map_gray_scaled;
  Convert(mk_disp_map,&disp_map_gray);
  ScaleFloatImage(mk_disp_map,&disp_map_gray_scaled);
  //ScaleDisparities(mk_disp_map, disparity_scale_factor, &disp_map_gray_scaled);
  char path_disp_map[200];
  char path_disp_map_scaled[200];
  sprintf(path_disp_map,"../../output/disp_map.pgm");
  sprintf(path_disp_map_scaled,"../../output/disp_map_scaled.pgm");
  Save(disp_map_gray,path_disp_map,"pgm");
  Save(disp_map_gray_scaled,path_disp_map_scaled,"pgm");

  //PredictOcclusion();

  end = clock();
  mk_total_time += (((double)(end - begin))/CLOCKS_PER_SEC);      
}

void MatchVNK::OverSegmentation()
{
  if(do_oversegm == true)
  {
    //Trace("into oversegmentation\n");

    ImgInt label_tmp = mk_label;
    std::vector<MatFlt> disp_func_tmp = mk_stb_disp_func;
    std::vector<float> centroid_x_tmp = mk_stb_centroid_x;
    std::vector<float> centroid_y_tmp = mk_stb_centroid_y;
  
    ImgInt label_change = mk_label;
    std::vector<MatFlt> disp_func_change = mk_stb_disp_func;
    std::vector<float> centroid_x_change = mk_stb_centroid_x;
    std::vector<float> centroid_y_change = mk_stb_centroid_y;

    mk_label = label_change;
    int e_prev = GetEnergy();
    int label_aff;
    label_aff = (skip_oversegm_label_zero ? 1 : 0);

    ImgFloat ref_img_smoothed, warp_img_smoothed, grad_x, grad_y;
    LKAPre(mk_ref_img,mk_warp_img,params,&ref_img_smoothed,&warp_img_smoothed,
           &grad_x,&grad_y);

    /*Figure fig1,fig2,fig3,fig4;
    fig1.Draw(ref_img_smoothed);
    fig2.Draw(warp_img_smoothed);
    fig3.Draw(grad_x);
    fig4.Draw(grad_y);
    AfxMessageBox("ok");*/

    clock_t start, finish;

    // debugBN
    //int i=0;
    //int j=3;
    //
    for(int i=label_aff; i<(disp_func_change.size()-1); i++)
      for(int j=(i+1); j<(disp_func_change.size()); j++)
      {
        //Trace(" i = %d j= %d size = %d\n",i,j,disp_func_change.size());
        if(IsNeighborCCImage(label_change,j,i))
        {
          //Trace(" i = %d j= %d are neighbors\n",i,j);
          ImgBinary mask;
          bool mask_flag = SetPixelsBinary(label_change,j,i,&mask);
          
          if(mask_flag) // if mask image has atleat one 'true' pixel
          {
            mk_label = label_change;
            mk_stb_disp_func = disp_func_change;
            mk_stb_centroid_x = centroid_x_change;
            mk_stb_centroid_y = centroid_y_change;

            MatFlt A_init(2,2);
            MatFlt d_init(1,2);

            A_init(0,0) = (mk_stb_disp_func[i])(0,0);
            A_init(1,0) = (mk_stb_disp_func[i])(1,0);
            d_init(0,0) = (mk_stb_disp_func[i])(2,0);
            A_init(0,1) = (mk_stb_disp_func[i])(3,0);
            A_init(1,1) = (mk_stb_disp_func[i])(4,0);
            d_init(0,1) = (mk_stb_disp_func[i])(5,0);

            start = clock();
            LKAResult result_change;
            params.skip_label_zero = true;
            //LKA(mk_ref_img,mk_warp_img,mask,params,A_init,d_init,&result_change);
            LKAPost(ref_img_smoothed,warp_img_smoothed,grad_x,grad_y,mask,
                    params,A_init,d_init,&result_change);
            //Trace("status after replacing %d with %d is %d\n",j,i,result_change.status);

            finish = clock();
            //Trace("\nTime taken (seconds): %lf\n",((double)(finish - start))/CLOCKS_PER_SEC);
            start = clock();
            if(result_change.status==0) // if tracked
            {
              //Trace("energy before replacing %d with %d is %d\n",j,i,e_prev);

              ReplacePixels(label_change,j,i,&mk_label);
  
              MatFlt z_change(6,1);
              z_change(0,0) = result_change.A(0,0);
              z_change(1,0) = result_change.A(1,0);
              z_change(2,0) = result_change.d(0,0);
              z_change(3,0) = result_change.A(0,1);
              z_change(4,0) = result_change.A(1,1);
              z_change(5,0) = result_change.d(0,1);
      
              mk_stb_disp_func[i] = z_change;
              mk_stb_centroid_x[i] = result_change.centroid_x;
              mk_stb_centroid_y[i] = result_change.centroid_y;

              mk_stb_disp_func[j] = z_change;
              mk_stb_centroid_x[j] = result_change.centroid_x;
              mk_stb_centroid_y[j] = result_change.centroid_y;

              ImgInt data_cost_img_tmp = mk_data_cost_img;           
              GetChangedEnergy(j,i);
              //GetEnergy();
              //Trace("energy after replacing %d with %d is %d\n",j,i,E);
              finish = clock();
              //Trace("\nTime taken (seconds): %lf\n",((double)(finish - start))/CLOCKS_PER_SEC);
              start = clock(); 
              if(E < e_prev)
              {
                //Trace("replacing label %d with %d\n",j,i);
           
                disp_func_change[i] = z_change;
                centroid_x_change[i] = result_change.centroid_x;
                centroid_y_change[i] = result_change.centroid_y;
              
                label_change = mk_label;
                for(int k=(j+1); k<(disp_func_change.size()); k++)
                {
                  ReplacePixels(label_change,k,k-1,&mk_label);
                  label_change = mk_label;
                
                  disp_func_change[k-1] = disp_func_change[k];
                  centroid_x_change[k-1] = centroid_x_change[k];
                  centroid_y_change[k-1] = centroid_y_change[k];              
                }
                disp_func_change.resize(disp_func_change.size()-1);
                centroid_x_change.resize(centroid_x_change.size()-1);
                centroid_y_change.resize(centroid_y_change.size()-1);

                j--;
                e_prev = E;

                //Trace("changed labels\n");
                finish = clock();
                //Trace("\nTime taken (seconds): %lf\n",((double)(finish - start))/CLOCKS_PER_SEC);

              }
              else
                mk_data_cost_img = data_cost_img_tmp;
            }
          }
        }
      }
  
    /*
    mk_label = label_tmp;
    mk_stb_disp_func = disp_func_tmp;
    mk_stb_centroid_x = centroid_x_tmp;
    mk_stb_centroid_y = centroid_y_tmp;
    */

    mk_label = label_change;
    mk_stb_disp_func = disp_func_change;
    mk_stb_centroid_x = centroid_x_change;
    mk_stb_centroid_y = centroid_y_change;

    //Trace("size of disp_func_change is %d\n",disp_func_change.size());

    // debug5 file - oversegmentation step finds equivalent labels based on energy
    std::vector<int> sts;
    sts.resize(disp_func_change.size(),9);
    char* debug_file_path5 = "../../output/debug5.txt";
    //WriteResults(disp_func_change,centroid_x_change,centroid_y_change,sts,debug_file_path5);
    //Trace("labels after over segmentation step\n");
    //TraceDispFunc(mk_stb_disp_func,mk_stb_centroid_x,mk_stb_centroid_y);
    //
  }
  
}

void MatchVNK::ResetLabelImage()
{
  mk_label.Reset(mk_width, mk_height);
  ImgInt::Iterator lbl_data_p;
	for(lbl_data_p=mk_label.Begin(); lbl_data_p!=mk_label.End(); lbl_data_p++) 
		*lbl_data_p = 0; // default label is 0
}

void MatchVNK::InitLabels()
{

  MatFlt z(6,1);
  Set(&z,0.0f);
  z(0,0) = 1.0f;
  z(4,0) = 1.0f;

  /*int i,j;
  for(i=mk_disp_x_min; i<=mk_disp_x_max; i++)
    for(j=mk_disp_y_min; j<=mk_disp_y_max; j++)
    {
      z(2,0) = (float) i;
      z(5,0) = (float) j;
      mk_stb_disp_func.push_back(z);
      mk_stb_centroid_x.push_back(0.0f);
      mk_stb_centroid_y.push_back(0.0f);
    }*/

  int i,j;
  for(j=mk_disp_y_min; j<=mk_disp_y_max; j++)
    for(i=mk_disp_x_min; i<=mk_disp_x_max; i++)
    {
      z(2,0) = (float) i;
      z(5,0) = (float) j;
      mk_stb_disp_func.push_back(z);
      mk_stb_centroid_x.push_back(0.0f);
      mk_stb_centroid_y.push_back(0.0f);
    }

  for(j=mk_disp_y_min; j<=mk_disp_y_max; j++)
    for(i=mk_disp_x_min; i<=mk_disp_x_max; i++)
    {
      mk_disp_x.push_back(i);
      mk_disp_y.push_back(j);
    }
}

void MatchVNK::MSegment()
{
	int i,j;
	int iter = multiway_cut_iter_num; // number of iterations
	int E_old;

	int num_labels = mk_stb_disp_func.size();
  //Trace("number of labels is %d\n",num_labels);
	int *label_buf; // flag for indication of successful energy reduction
	label_buf = new int[num_labels]; 

	int buf_num=num_labels; // number of labels active
	int label;

	for(label=0;label<num_labels;label++)
		label_buf[label] = 0; // initially all label expansions can reduce energy

  E = GetEnergy();

  clock_t startbig, finishbig;
  startbig = clock();

	for(i=0;(i<iter)&&(buf_num>0);i++)
	{
		for(label=0;label<num_labels;label++)
		{
      E_old = E;
      clock_t start, finish;
      start = clock();
      AExpand(label);
      finish = clock();
      //Trace("Time for label %d expansion (seconds): %lf\n",label,((double)(finish - start))/CLOCKS_PER_SEC);
      		
			if (E_old == E)
			{
				if (!label_buf[label]) { label_buf[label] = 1; buf_num --; }
			}
			else
			{
				for (j=0; j<num_labels; j++) label_buf[j] = 0;
				label_buf[label] = 1;
				buf_num = num_labels - 1;
			}
		}
	}

  delete[] label_buf;

  finishbig = clock();

  ////Trace("\nenergy after multiway cuts %d\n",E);

  /*char buffer[200];
  sprintf(buffer,"Multiway cut Time taken (seconds): %lf",((double)(finishbig - startbig))/CLOCKS_PER_SEC);
  AfxMessageBox(buffer);*/

}

int MatchVNK::GetEnergy()
{
	int k;
	Location p, q;
	int p_lbl,q_lbl;
	
	E = 0;

  int data_tmp;
  for(p.t=0; p.t<=0; p.t++)
    for(p.y=0; p.y<mk_height; p.y++)
      for(p.x=0; p.x<mk_width; p.x++)
			{
				p_lbl = mk_label(p.x, p.y);
        
        data_tmp = ((this->*DataCost)(p,p_lbl));
        mk_data_cost_img(p.x, p.y) = data_tmp;
				E += data_tmp;

				for (k=0; k<4; k++)
				{
					q = p + N[k];

					if (q>=PS && q<=PE)
					{
						q_lbl = mk_label(q.x, q.y);
            E += SmoothnessCost(p,p_lbl,q,q_lbl);
					}
				}
      }
 
	return E;
}

int MatchVNK::GetChangedEnergy(const int label1, const int label2)
{
  int k;
	Location p, q;
	int p_lbl,q_lbl;
	
	E = 0;

  int data_tmp;
  for(p.t=0; p.t<=0; p.t++)
    for(p.y=0; p.y<mk_height; p.y++)
      for(p.x=0; p.x<mk_width; p.x++)
			{
				p_lbl = mk_label(p.x, p.y);

        if(p_lbl == label1 || p_lbl == label2)
        {
          data_tmp = ((this->*DataCost)(p,p_lbl));
          mk_data_cost_img(p.x, p.y) = data_tmp;
				  E += data_tmp;
        }
        else
          E += mk_data_cost_img(p.x, p.y);

				for (k=0; k<4; k++)
				{
					q = p + N[k];

					if (q>=PS && q<=PE)
					{
						q_lbl = mk_label(q.x, q.y);
            E += SmoothnessCost(p,p_lbl,q,q_lbl);
					}
				}
			}
 
	return E;
}

void energy_error_func(char *msg)
{
	AfxMessageBox(msg);
	assert(0);
}

#define VAR_ACTIVE ((Energy::Var)0)

#define BVZ_ALPHA_SINK
/*
	if BVZ_ALPHA_SINK is defined then interpretation of a cut is as follows:
		SOURCE means initial label
		SINK   means new label \alpha

	if BVZ_ALPHA_SINK is not defined then SOURCE and SINK are swapped
*/
#ifdef BVZ_ALPHA_SINK
	#define ADD_TERM1(var, E0, E1) add_term1(var, E0, E1)
	#define ADD_TERM2(var1, var2, E00, E01, E10, E11) add_term2(var1, var2, E00, E01, E10, E11)
	#define VALUE0 0
	#define VALUE1 1
#else
	#define ADD_TERM1(var, E0, E1) add_term1(var, E1, E0)
	#define ADD_TERM2(var1, var2, E00, E01, E10, E11) add_term2(var1, var2, E11, E10, E01, E00)
	#define VALUE0 1
	#define VALUE1 0
#endif


void MatchVNK::AExpand(int label)
{
	Location p, q;
	int p_lbl, q_lbl;

	Energy::Var p_var, q_var;
	int E_old, E00, E0a, Ea0;
	int k;

	Energy *e = new Energy(energy_error_func);
	assert(e);

  // initializing
  for (p.t=0; p.t<=0; p.t++)
	{
	  for (p.y=0; p.y<mk_height; p.y++)
	    for (p.x=0; p.x<mk_width; p.x++)
	    {
		    p_lbl = (*(mk_label.Begin(p.x, p.y)));
        		
		    if (label == p_lbl)
		    {
			    *(mk_ptr_img + ((p.y)*mk_width + p.x)) = VAR_ACTIVE;
          e -> add_constant(((this->*DataCost)(p,p_lbl)));
        }
		    else
		    {
          p_var = e -> add_variable();
			    assert(p_var);
			
          *(mk_ptr_img + ((p.y)*mk_width + p.x)) = p_var;
          e -> ADD_TERM1(p_var, ((this->*DataCost)(p,p_lbl)), ((this->*DataCost)(p, label)));
		    }
	    }
	}

	for (p.t=0; p.t<=0; p.t++)
	{
	  for (p.y=0; p.y<mk_height; p.y++)
	  {
	    for (p.x=0; p.x<mk_width; p.x++)
	    {
		    p_lbl = (*(mk_label.Begin(p.x, p.y)));
		    p_var = (Energy::Var) (*(mk_ptr_img + ((p.y)*mk_width + p.x)));

		    // smoothness term
		    for (k=0; k<4; k++)
		    {
			    q = p + N[k];

			    if ( ! (q>=PS && q<=PE) ) continue;

			    q_var = (Energy::Var) (*(mk_ptr_img + ((q.y)*mk_width + q.x)));
			    q_lbl = (*(mk_label.Begin(q.x, q.y)));

			    if (p_var != VAR_ACTIVE && q_var != VAR_ACTIVE)
				  E00 = SmoothnessCost(p, p_lbl, q, q_lbl);
			    if (p_var != VAR_ACTIVE)
				  E0a = SmoothnessCost(p, p_lbl, q, label);
			    if (q_var != VAR_ACTIVE)
				  Ea0 = SmoothnessCost(p, label, q, q_lbl);

			    if (p_var != VAR_ACTIVE)
			    {
				    if (q_var != VAR_ACTIVE) e -> ADD_TERM2(p_var, q_var, E00, E0a, Ea0, 0); 
            else e -> ADD_TERM1(p_var, E0a, 0);
          }
			    else
			    {
				    if (q_var != VAR_ACTIVE) e -> ADD_TERM1(q_var, Ea0, 0);
				    else {}
			    }
	      }
	    }
    }
	}


	E_old = E;
	E = e -> minimize();

	if (E < E_old)
	{
		for (p.t=0; p.t<=0; p.t++)
		  for (p.y=0; p.y<mk_height; p.y++)
		    for (p.x=0; p.x<mk_width; p.x++)
		    {
			    p_var = (Energy::Var) (*(mk_ptr_img + ((p.y)*mk_width + p.x)));
			    if (p_var!=VAR_ACTIVE && e->get_var(p_var)==VALUE1)
			    {
				    *(mk_label.Begin(p.x, p.y)) = label;
			    }
		    }
	}

	delete e;
}

int __cdecl MatchVNK::DataCostInit(Location p, int p_lbl)
{

  Location p_disp;
  p_disp.x = (int) (((float) p.x) + (*((mk_stb_disp_func[p_lbl]).Begin(2,0))));
  p_disp.y = (int) (((float) p.y) + (*((mk_stb_disp_func[p_lbl]).Begin(5,0))));
  p_disp.t = 0;

  if(p_disp>=PS && p_disp<=PE)
  {
	  float dl, dr, d;
	  float Il, Il_min, Il_max, Ir, Ir_min, Ir_max;

	  Il     = (*(mk_ref_img.Begin(p.x,p.y))); 
	  Il_min = (*(mk_ref_min_img.Begin(p.x,p.y))); 
	  Il_max = (*(mk_ref_max_img.Begin(p.x,p.y))); 

    Ir     = (*(mk_warp_img.Begin(p_disp.x,p_disp.y)));
    Ir_min = (*(mk_warp_min_img.Begin(p_disp.x,p_disp.y)));
    Ir_max = (*(mk_warp_max_img.Begin(p_disp.x,p_disp.y)));

	  if      (Il < Ir_min) dl = Ir_min - Il;
	  else if (Il > Ir_max) dl = Il - Ir_max;
	  else return 0;

	  if      (Ir < Il_min) dr = Il_min - Ir;
	  else if (Ir > Il_max) dr = Ir - Il_max;
	  else return 0;

    if(dl < dr)
      d = dl;
    else
      d = dr;

	  /*d = d*d;
	  if (d > data_cost_limit) d = data_cost_limit;
    return ((int) d);*/
	  d = fabsf(d);
	  if (d > data_cost_limit) d = data_cost_limit;
    return ((int) d);
  }
  else
  {
    return ((int) data_cost_oob);
  }
  
}

#ifdef HARD_CONSTRAINTS

int __cdecl MatchVNK::DataCostIter(Location p, int p_lbl)
{

  A(0,0) = *((mk_stb_disp_func[p_lbl]).Begin(0,0));
  A(1,0) = *((mk_stb_disp_func[p_lbl]).Begin(1,0));
  A(0,1) = *((mk_stb_disp_func[p_lbl]).Begin(3,0));
  A(1,1) = *((mk_stb_disp_func[p_lbl]).Begin(4,0));

  d(0,0) = *((mk_stb_disp_func[p_lbl]).Begin(2,0));
  d(0,1) = *((mk_stb_disp_func[p_lbl]).Begin(5,0));

  centroid(0,0) = mk_stb_centroid_x[p_lbl];
  centroid(0,1) = mk_stb_centroid_y[p_lbl];

  loc_in(0,0) = ((float) p.x) - centroid(0,0);
  loc_in(0,1) = ((float) p.y) - centroid(0,1);
  loc_out = (A*loc_in) + d + centroid;

  Location p_disp;
  p_disp.x = (int) (loc_out(0,0));
  p_disp.y = (int) (loc_out(0,1));
  p_disp.t = 0;


    if(p_disp>=PS && p_disp<=PE)
    {
      float d;
      d = (*(mk_ref_img.Begin(p.x,p.y))) - Interp(mk_warp_img,loc_out(0,0),loc_out(0,1));

      if(mk_corr_label(p.x,p.y) != 0 && 
         fabsf(fabsf(loc_out(0,0)-p.x)-fabsf(mk_corr_label(p.x,p.y))) > 0.5)
        d = d*hard_penalty_factor;

      //if(mk_hard_constraints(p.x,p.y) != 0 && 
         //fabsf(fabsf(loc_out(0,0)-p.x)-fabsf(mk_hard_constraints(p.x,p.y))) > 0.5)
        //d = d*hard_penalty_factor;

      /*d = d*d;
	    if (d > data_cost_limit) d = data_cost_limit;
      return ((int) d);*/
	    d = fabsf(d);
	    if (d > data_cost_limit) d = data_cost_limit;
      return ((int) d);
    }
    else
    {
      return ((int) data_cost_oob);
    }
}

#else

int __cdecl MatchVNK::DataCostIter(Location p, int p_lbl)
{

  A(0,0) = *((mk_stb_disp_func[p_lbl]).Begin(0,0));
  A(1,0) = *((mk_stb_disp_func[p_lbl]).Begin(1,0));
  A(0,1) = *((mk_stb_disp_func[p_lbl]).Begin(3,0));
  A(1,1) = *((mk_stb_disp_func[p_lbl]).Begin(4,0));

  d(0,0) = *((mk_stb_disp_func[p_lbl]).Begin(2,0));
  d(0,1) = *((mk_stb_disp_func[p_lbl]).Begin(5,0));

  centroid(0,0) = mk_stb_centroid_x[p_lbl];
  centroid(0,1) = mk_stb_centroid_y[p_lbl];

  loc_in(0,0) = ((float) p.x) - centroid(0,0);
  loc_in(0,1) = ((float) p.y) - centroid(0,1);
  loc_out = (A*loc_in) + d + centroid;

  Location p_disp;
  p_disp.x = (int) (loc_out(0,0));
  p_disp.y = (int) (loc_out(0,1));
  p_disp.t = 0;


    if(p_disp>=PS && p_disp<=PE)
    {
      float d;
      d = (*(mk_ref_img.Begin(p.x,p.y))) - Interp(mk_warp_img,loc_out(0,0),loc_out(0,1));

      /*d = d*d;
	    if (d > data_cost_limit) d = data_cost_limit;
      return ((int) d);*/
	    d = fabsf(d);
	    if (d > data_cost_limit) d = data_cost_limit;
      return ((int) d);
    }

    /*if(p_disp>=PS && p_disp<=PE)
    {
	    float dl, dr, d;
	    float Il, Il_min, Il_max, Ir, Ir_min, Ir_max;

	    Il     = (*(mk_ref_img.Begin(p.x,p.y))); 
	    Il_min = (*(mk_ref_min_img.Begin(p.x,p.y))); 
	    Il_max = (*(mk_ref_max_img.Begin(p.x,p.y))); 

      Ir     = (*(mk_warp_img.Begin(p_disp.x,p_disp.y)));
      Ir_min = (*(mk_warp_min_img.Begin(p_disp.x,p_disp.y)));
      Ir_max = (*(mk_warp_max_img.Begin(p_disp.x,p_disp.y)));

	    if      (Il < Ir_min) dl = Ir_min - Il;
	    else if (Il > Ir_max) dl = Il - Ir_max;
	    else return 0;

	    if      (Ir < Il_min) dr = Il_min - Ir;
	    else if (Ir > Il_max) dr = Ir - Il_max;
	    else return 0;

      if(dl < dr)
        d = dl;
      else
        d = dr;

	    d = d*d;
	    if (d > data_cost_limit) d = data_cost_limit;
      return ((int) d);
	    //d = fabsf(d);
	    //if (d > data_cost_limit) d = data_cost_limit;
      //return ((int) d);
    }*/

    else
    {
      return ((int) data_cost_oob);
    }
}

#endif


#ifdef CANNY_SMOOTHNESS

int MatchVNK::SmoothnessCost(Location p, int p_lbl, Location q, int q_lbl)
{
  if( p_lbl == q_lbl)
    return 0;
  else
  {
    float edge_diff = (float) (mk_canny_edge(p.x,p.y) - mk_canny_edge(q.x,q.y));
    if(fabsf(edge_diff) < 1.0f)
      return lambda1;
    else
      return lambda2;
  }
}

#else

int MatchVNK::SmoothnessCost(Location p, int p_lbl, Location q, int q_lbl)
{
  if( p_lbl == q_lbl)
    return 0;
  else
  {
    float int_diff = (*(mk_ref_img.Begin(p.x,p.y))) - (*(mk_ref_img.Begin(q.x,q.y)));
    if(int_diff < 0) int_diff = -int_diff;

    if(int_diff < I_threshold)
      return lambda1;
    else
      return lambda2;
  }

}

#endif

void MatchVNK::CreateDispMap()
{
  mk_disp_map.Reset(mk_width, mk_height);
  ImgInt::ConstIterator lbl_p;
  ImgFloat::Iterator dm_p;
  int i=0,j=0;
  for(lbl_p = mk_label.Begin(), dm_p = mk_disp_map.Begin();
      lbl_p != mk_label.End(); lbl_p++, dm_p++)
  {
    A(0,0) = *((mk_stb_disp_func[*lbl_p]).Begin(0,0));
    A(1,0) = *((mk_stb_disp_func[*lbl_p]).Begin(1,0));
    A(0,1) = *((mk_stb_disp_func[*lbl_p]).Begin(3,0));
    A(1,1) = *((mk_stb_disp_func[*lbl_p]).Begin(4,0));

    d(0,0) = *((mk_stb_disp_func[*lbl_p]).Begin(2,0));
    d(0,1) = *((mk_stb_disp_func[*lbl_p]).Begin(5,0));

    centroid(0,0) = mk_stb_centroid_x[*lbl_p];
    centroid(0,1) = mk_stb_centroid_y[*lbl_p];

    loc_in(0,0) = ((float) i) - centroid(0,0);
    loc_in(0,1) = ((float) j) - centroid(0,1);
    loc_out = (A*loc_in) + d + centroid;
    
    //*dm_p = fabsf((float)i - loc_out(0,0));
    *dm_p = sqrtf((fabsf((float)i - loc_out(0,0)))*(fabsf((float)i - loc_out(0,0))) +
                  (fabsf((float)j - loc_out(0,1)))*(fabsf((float)j - loc_out(0,1))));

    i++;
    if(i == mk_label.Width())
    {
      i=0;
      j++;
    }
  }

  Figure fig;
  fig.Draw(mk_disp_map);
  //AfxMessageBox("ok");
}

void MatchVNK::AdjustBorder()
{
  int i,j;
  for(i=border_adjust_width; i>=0; i--)
    for(j=0; j<mk_height; j++)
      mk_label(i,j) = mk_label(i+1,j);
}